#pragma once
#include "cutil.h"
#include "types.h"
#include "problem.h"

#include <string>
#include <sstream>

using std::string;
using std::stringstream;

class Benchmark {
protected:
  Problem problem_;
  unsigned int timer_;
  FLOAT elapsed_time_;
  FLOAT result_time_;
  FLOAT result_accuracy_;
  FLOAT result_flops_;

  inline void startTimer() {
    CUT_SAFE_CALL(cutCreateTimer(&timer_));
    CUT_SAFE_CALL(cutResetTimer(timer_));
    CUT_SAFE_CALL(cutStartTimer(timer_));
  }
  
  inline void stopTimer() {
    CUT_SAFE_CALL(cutStopTimer(timer_));
    elapsed_time_ = (FLOAT)cutGetTimerValue(timer_) * 1.0e-03;
    CUT_SAFE_CALL(cutDeleteTimer(timer_));
  }

public:
  Benchmark() {}
  ~Benchmark() {}

  inline
  const Problem& problem() const { return problem_; }
  virtual void initialize(const int nx, const int ny, const int nz) {
    problem_.initialize(nx, ny, nz);
  }
  virtual void run() = 0;
  virtual string result() const {
    stringstream ss;
    ss << "Elapsed time    : " << elapsed_time_ << "[sec]\n"
       << "Time calculated : " << result_time_ << "\n"
       << "Accuracy        : " << result_accuracy_ << "\n"
       << "Performance     : " << result_flops_ << "[MFlops]\n";
    return ss.str();
  }
};
