#include <cuda/cuda.h>
#include <cutil.h>
#include <iostream>
#include <string>
#include <algorithm>
#include <cmath>
#include "log.h"
#include "cpu_benchmark.h"
#include "gpu_benchmark.h"
#include "omp_benchmark.h"

using std::cout;
using std::string;

enum Target {
    CPU, GPU, OMP, MULTI_GPU
};

static int flag_size = 64;
static Target flag_target = CPU;

static void usage() {
  std::cout << "usage: bench [-cpu|-gpu] [-size size]\n";
  exit(EXIT_FAILURE);
}

static void parseFlags(int argc, char **argv) {
  for (int i = 1; i < argc; i++) {
    string arg(argv[i]);
    if (arg == "-cpu") {
      flag_target = CPU;
    }
    else if (arg == "-gpu") {
      flag_target = GPU;
    }
    else if (arg == "-size") {
      if (++i >= argc) {
        usage();
      }
      flag_size = atoi(argv[i]);
    }
    else if (arg == "-openmp") {
        flag_target = OMP;        
    }
    else if (arg == "-multi-gpu") {
        flag_target = MULTI_GPU;
    }
    else {
      usage();
    }
  }
}

int main(int argc, char **argv) {
  parseFlags(argc, argv);

  Benchmark *benchmark;

  switch (flag_target) {
  case CPU:
      benchmark = new CPUBenchmark();
      break;
  case GPU:
      benchmark = new GPUBenchmark();
      break;
  case OMP:
      benchmark = new OMPBenchmark();
      break;
  case MULTI_GPU:
      benchmark = new OMPGPUBenchmark();
      break;
  }

  benchmark->initialize(flag_size, flag_size, flag_size);
  LOG_INFO("\n%s", benchmark->problem().debug_string().c_str());
  benchmark->run();
  LOG_INFO("\n%s", benchmark->result().c_str());

  return EXIT_SUCCESS;
}
