#include <iostream>
#include <string.h>
#include "omp_benchmark.h"
#include "log.h"

using namespace std;

void OMPBenchmark::initialize(const int nx, const int ny, const int nz) {
    problem_.initialize(nx, ny, nz);
}

void OMPBenchmark::run() {
    int    nx        = problem_.n.x;
    int    ny        = problem_.n.y;
    int    nz        = problem_.n.z;
    int    size      = nx * ny * nz;
    FLOAT  dx        = problem_.d.x;
    FLOAT  dy        = problem_.d.y;
    FLOAT  dz        = problem_.d.z;
    FLOAT  dt        = problem_.dt;
    FLOAT  k         = problem_.kappa;
    FLOAT *f         = new FLOAT[size];
    FLOAT *fn        = new FLOAT[size];
    memcpy(f, problem_.f(), size * sizeof(FLOAT));
    FLOAT final_time;
    int final_iter_count;

    LOG_INFO("starting calculation");

    startTimer();

#pragma omp parallel 
    {
        FLOAT  time = 0.0;
        int    iter_count = 0;
        int    tid       = omp_get_thread_num();
        int    nthreads  = omp_get_num_threads();
        int    dim_z     = nz / nthreads;
        int    rem_z     = nz % nthreads;
        int    nz_self   = tid == 0 ? dim_z + rem_z : dim_z;
        if (tid == 0) LOG_INFO("%d threads running", nthreads);
        do {
            if (tid == 0 && iter_count && (iter_count % 100 == 0)) {
                LOG_INFO("time(%4d)=%7.5f", iter_count, time + dt);
            }
            diffusion3d(f, fn, nx, ny, nz_self, dx, dy, dz, dt, k);
#pragma omp barrier 
            std::swap(f, fn);
            time += dt;
            iter_count++;
        } while (time + 0.5*dt < 0.1);
        if (tid == 0) {
            final_time = time;
            final_iter_count = iter_count;
        }
    }
    
    stopTimer();
    LOG_INFO("calculation has been done.\n");
    result_time_     = final_time;
    result_accuracy_ = problem_.accuracy(fn, final_time);
    result_flops_    = (nx*ny*nz)*13.0*final_iter_count/elapsed_time_*1.0e-06;

    delete[] f;
    delete[] fn;
}
