#pragma once
#include "benchmark.h"
#include <omp.h>

class OMPBenchmark : public Benchmark {
protected:
    FLOAT   *fore_buffer;
    FLOAT   *back_buffer;

    inline void diffusion3d(
            FLOAT *f, FLOAT *fn, int nx, int ny, int nz,
            FLOAT dx, FLOAT dy, FLOAT dz, FLOAT dt, FLOAT kappa) {
        FLOAT ce, cw, cn, cs, ct, cb, cc;
        ce = cw = kappa*dt/(dx*dx);
        cn = cs = kappa*dt/(dy*dy);
        ct = cb = kappa*dt/(dz*dz);
        cc = 1.0 - (ce + cw + cn + cs + ct + cb);

        int j, jx, jy, jz;
        for (jz = 0; jz < nz; jz++) {
            for (jy = 0; jy < ny; jy++) {
                for (jx = 0; jx < nx; jx++) {
                    FLOAT e, w, n, s, t, b, c;
                    j = (jz+1)*nx*ny + jy*nx + jx;
                    c = f[j];
                    w = (jx == 0)    ? c : f[j-1];
                    e = (jx == nx-1) ? c : f[j+1];
                    n = (jy == 0)    ? c : f[j-nx];
                    s = (jy == ny-1) ? c : f[j+nx];
                    b = (jz == 0)    ? c : f[j-nx*ny];
                    t = (jz == nz-1) ? c : f[j+nx*ny];
                    fn[j] = cc*c
                            + cw*w + ce*e
                            + cs*s + cn*n
                            + cb*b + ct*t;
                }
            }
        }
    }

public:
    OMPBenchmark() : fore_buffer(NULL), back_buffer(NULL) {
    }

    virtual ~OMPBenchmark() {
        delete[] fore_buffer;
        delete[] back_buffer;
    }

    virtual void initialize(const int nx, const int ny, const int nz);
    virtual void run();
};


class OMPGPUBenchmark : public OMPBenchmark {
private:
public:
    OMPGPUBenchmark() {}
    virtual void run();
};

