#include "problem.h"
//#include <cutil.h>
#include <cmath>
#include <string>
#include <sstream>

using std::string;
using std::stringstream;

string Problem::debug_string() const {
  stringstream ss;
  ss << "Problem size: (" << n << ")\n"
     << "l           : (" << l << ")\n"
     << "d           : (" << d <<  ")\n"
     << "k           : (" << k <<  ")\n"
     << "kappa       : "  << kappa << "\n"
     << "dt          : "  << dt << "\n";
  return ss.str();
}

void Problem::initialize(const int nx, const int ny, const int nz) {
  delete[] f_;
  n     = Size3d(nx, ny, nz);
  l     = Point3df(1.0     , 1.0     , 1.0     );
  d     = Point3df(l.x/n.x , l.y/n.y , l.z/n.z );
  k     = Point3df(2.0*M_PI, 2.0*M_PI, 2.0*M_PI);
  kappa = 0.1;
  dt    = 0.1*d.x*d.x / kappa;
  f_    = new FLOAT[nx*ny*nz];

  for (int jz = 0; jz < nz; jz++) {
    for (int jy = 0; jy < ny; jy++) {
      for (int jx = 0; jx < nx; jx++) {
        int j = index(jx, jy, jz);
        FLOAT x = d.x*((FLOAT)(jx + 0.5));
        FLOAT y = d.y*((FLOAT)(jy + 0.5));
        FLOAT z = d.z*((FLOAT)(jz + 0.5));
        f_[j] = (FLOAT)0.125
          * (1.0 - cos(k.x*x))
          * (1.0 - cos(k.y*y))
          * (1.0 - cos(k.z*z));
      }
    }
  }
}

FLOAT Problem::accuracy(const FLOAT *f, const FLOAT time) const {
  FLOAT ferr = 0.0;
  Point3df a(exp(-kappa*time*(k.x*k.x)),
             exp(-kappa*time*(k.y*k.y)),
             exp(-kappa*time*(k.z*k.z)));
  for (int jz = 0; jz < n.z; jz++) {
    for (int jy = 0; jy < n.y; jy++) {
      for (int jx = 0; jx < n.x; jx++) {
        int j = index(jx, jy, jz);
        FLOAT x = d.x*((FLOAT)(jx + 0.5));
        FLOAT y = d.y*((FLOAT)(jy + 0.5));
        FLOAT z = d.z*((FLOAT)(jz + 0.5));

        FLOAT f0 = (FLOAT)0.125
          *(1.0 - a.x*cos(k.x*x))
          *(1.0 - a.y*cos(k.y*y))
          *(1.0 - a.z*cos(k.z*z));
        ferr += (f[j] - f0)*(f[j] - f0);
      }
    }
  }
  return (FLOAT)sqrt(ferr/(n.x*n.y*n.z));
}
