/*
 * Based on the code by Toshio Endo <endo@is.titech.ac.jp>
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

long time_diff_us(struct timeval st, struct timeval et)
{
    return (et.tv_sec-st.tv_sec)*1000000+(et.tv_usec-st.tv_usec);
}

int initialize(float *region[2], int nx, int ny)
{
    int rid;
    int i, j;
    /* initialize two regions */
    for (rid = 0; rid < 2; rid++) {
	
        for (i = 0; i < ny; i++) {
            for (j = 0; j < nx; j++) {
                region[rid][i*nx+j] = 0.0;
            }
        }
        /* right edge */
        for (i = 0; i < ny; i++) {
            region[rid][i*nx+(nx-1)] = (float)i/(float)ny;
        }
        /* lower edge */
        for (j = 0; j < nx; j++) {
            region[rid][(ny-1)*nx+j] = (float)j/(float)nx;
        }
    }
    return 0;
}

void sor(float *region[2], int nx, int ny, float omega)
{
    int i, j;
    int iter;
    int max_iter = 10000;    
    /* buffer id */
    int cur = 0;
    int next = 1;

    for (iter = 0; iter < max_iter; iter++) {
        for (i = 1; i < ny-1; i++) {
            for (j = 1; j < nx-1; j++) {
                float curv = region[cur][i*nx+j];
                float nextv;
                float onextv;
                float resid;
                /* compute new value from neighbor points */
                nextv = (region[cur][(i-1)*nx+j] +
                         region[cur][(i+1)*nx+j] +
                         region[cur][i*nx+(j-1)] +
                         region[cur][i*nx+(j+1)]) * 0.25f;

                /* overcorrection */
                onextv = curv + (nextv - curv) * omega;
                region[next][i*nx+j] = onextv;
            }
        }

        /* switch buffer */
        cur = 1-cur;
        next = 1-next;
    }

    return;
}

int main(int argc, char *argv[])
{
    int nx, ny;
    float omega;
    float *region[2]; /* for double buffer */
    int i, j;
    int rid;
    struct timeval st;
    struct timeval et;
    long us;
    int iter;

    if (argc < 4) {
        printf("Specify nx, ny, omega.\n");
        exit(1);
    }

    nx = atoi(argv[1]);
    ny = atoi(argv[2]);
    omega = atof(argv[3]);

    /* allocate 2D region (row major) */
    /* double buffer */
    region[0] = (float*)malloc(sizeof(float)*nx*ny);
    region[1] = (float*)malloc(sizeof(float)*nx*ny);


    initialize(region, nx, ny);

    gettimeofday(&st, NULL); /* get start time */
    sor(region, nx, ny, omega);
    gettimeofday(&et, NULL); /* get end time */

    us = time_diff_us(st, et);
    printf("SOR took %ld us, %d iterations --> %lf MFlops\n",
           us, iter, (double)iter*(double)(nx*ny*4)/(double)us);

    free(region[0]);
    free(region[1]);
    return 0;
}
